/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  System       : EEPROM emulation
  Module name  : EEPROM emulation control processing
  File name    : EEPROMCtrlEx.c
  Version name : 1.00
  Target CPU   : uPD78F0547(78K0/KF2)
  Last updated : 2005/09/08
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#pragma section @@CODE	SLF_CTRL

#include "EEPROMCtrlEx.h"


/*-------------------------------------------------------------------
 Function name : ucEEPROMInitEx(user access function)
 Arguments     : None
 Return value  : Normal(TRUE:00H) / Noemal end(Block error)(NMLBLK:01H) /
                 Prohibition of all block use(ERRALL:FBH) / Device error(ERRDEV:FDH) / 
                 Initialization error(ERRINIT:FEH)
 Summary       : initialize EEPROM emulation
------------------------------------------------------------------*/
UCHAR ucEEPROMInitEx( void )										/* EEPROM emulation Initialize processing */
{
	USHORT	usFlagAdr;			/* Effective/invalid flag address */
	UCHAR	ucFindBlock;		/* Discovery block num ber        */
	UCHAR	ucCount;			/* Elimination counter            */
	UCHAR	ucFlags;			/* Various flags                  */
								/*   D7:Block error               */
								/*   D6:Block elimination         */
								/*   D5:not use                   */
								/*   D4:not use                   */
								/*   D3:not use                   */
								/*   D2:not use                   */
								/*   D1:not use                   */
								/*   D0:not use                   */
	UCHAR	ucResult;			/* Processing result              */
	UCHAR	i;
	
	/* Parameter Check */
	if( PARAM_OK == FALSE ){										/* parameter error? */
		return( ERRPRM );											/* then rerun */
	}
	
	/* Variable initialization */
	ucFindBlock = 0xFF;												/* Discovery block number initialization */
	ucFlags		= 0x00;												/* Various flag initialization */
	/* Block elimination and effective block setting processing */
	for ( i = 0 ; i < EEPROM_BLOCK_NO ; i++ ){						/* loop for use block several minutes */
		/* Elimination flag Off */
		ucFlags &= ~FL_DOERS;
		/* Bank setting */
		SET_BANK( ucEEPROM_BANK[ i ] );
		/* The bank top address (a flag storage address) acquisition */
		usFlagAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ i ] );
		/* An elimination condition judgment */
		if( BLOCK_IS_PROHIBIT( (UCHAR*)usFlagAdr ) == FALSE ){		/* A block is a ban on use state? */
			if ( BLOCK_IS_EFFECTIVE( (UCHAR*)usFlagAdr ) == TRUE ){	/* A block is effective? */
				if ( ucFindBlock == 0xFF ){							/* Effective block non-discovery */
					ucFindBlock = i;								/* Block number setting */
				}
				else{												/* finished with effective block discovery */
					ucFlags |= FL_DOERS;							/* Elimination flag On */
				}
			}
			else{													/* block=An invalid state (an abnormal state) */
				ucFlags |= FL_DOERS;								/* Elimination flag On */
			}
		}
		/* Processing to erase a block */
		if ( ucFlags & FL_DOERS ){									/* Elimination flag On? */
			SelfFlashModeOn();										/* changes to a self-programming mode */
			if ( ucSelfFlashInitialize() == FALSE ){				/* self-programming Initialization */
				SelfFlashModeOff();									/* changes to a normal mode */
				return ( ERRINIT );									/* An initialization error */
			}
			/* An elimination counter minute loop */
			for ( ucCount = ERASENUM + 1; ucCount > 0; ucCount-- ) {
				/* Erases data in the block next used as EEPROM */
				if ( ucSelfFlashBlockErase( ucEEPROM_BANK[ i ], ucEEPROM_BLOCK[ i ] ) == TRUE ){	/* I */
					break;											/* loop end */
				}
			}
			/* An elimination result judgment */
			if ( ucCount <= 0 ){									/* Count<=0 (re-try over) */
				/* make a block the ban on use */
				if ( ucEEPROMUnableEx( i ) == FALSE ){				/* Abnormal termination */
					SelfFlashModeOff();								/* changes to a normal mode */
					return ( ERRDEV );								/* device error */
				}
				ucFlags |= FL_BLKERR;								/* Block error flag On */
			}
			SelfFlashModeOff();										/* changes to a normal mode */
		}
	}
	
	/* Setting handling of effective blocking */
	ucResult = TRUE;												/* Return value setting */
	if ( ucFindBlock == 0xFF ){										/* Effective block non-discovery */
		for ( i = 0 ; i < EEPROM_BLOCK_NO ; i++ ){					/* loop for use block several minutes */
			/* Bank setting */
			SET_BANK( ucEEPROM_BANK[ i ] );
			/* The bank top address (a flag storage address) acquisition */
			usFlagAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ i ] );
			/* A flag state judgment */
			if( BLOCK_IS_BLANK( (UCHAR*)usFlagAdr ) == TRUE ){		/* A block is a state finished with elimination? */
				ucDataBuf[ 0 ] = 0x55;
				ucDataBuf[ 1 ] = 0x55;
				ucDataBuf[ 2 ] = 0x55;
				ucDataBuf[ 3 ] = 0x55;								/* Provids block validity setting data */
				SelfFlashModeOn();									/* changes to a self-programming mode */
				if ( ucSelfFlashInitialize() == FALSE ){			/* self-programming Initialization */
					SelfFlashModeOff();								/* changes to a normal mode */
					return ( ERRINIT );								/* initialization error */
				}
				if ( ucSelfFlashEEPROMWrite( ucEEPROM_BANK[ i ], usFlagAdr, 4-1 ) == TRUE ){	/* Sets block 0 as valid block */
					SelfFlashModeOff();								/* changes to a normal mode */
					break;
				}
				else {
					/* make a block the ban on use */
					if ( ucEEPROMUnableEx( i ) == FALSE ){			/* abnormal end */
						SelfFlashModeOff();							/* changes to a normal mode */
						ucResult = ERRDEV;							/* device error */
						break;
					}
					ucFlags |= FL_BLKERR;							/* Block error flag On */
				}
				SelfFlashModeOff();									/* changes to a normal mode */
			}
		}
		if ( i >= EEPROM_BLOCK_NO ){								/* not able to set a block effectively */
			ucResult = ERRALL;										/* Prohibition of all block use */
		}
	}
	if ( ucResult == TRUE ){										/* Normal end */
		if ( ucFlags & FL_BLKERR ){									/* Those with a block error */
			ucResult = NMLBLK;										/* Normal end(Block error generating) */
		}
	}
	return ( ucResult );
}

/*-------------------------------------------------------------------
 Function name : ucEEPROMReadEx(user access function)
 Arguments	   : ucDataNo:Read data No.
		       : *pDataAddress:Read data storage address
 Return value  : Normal completion(TRUE:00H) / Reading data nothing(ERRNODAT:F7H) / 
                 With no effective block(ERRNOUSE:FAH)
 Summary       : Data with the specified number is read from EEPROM to the storage address.
		       : 1.Searches for blocks being used as EEPROM.
		       : 2.Searches for latest data's address among valid blocks.
		       : 3.Reads latest data from searched address.
-------------------------------------------------------------------*/
UCHAR ucEEPROMReadEx( UCHAR ucDataNo, UCHAR *pDataAddress )			/* EEPROM data read processing */
{
	USHORT	usDataSrcAdr;
	UCHAR	ucBlockTBLNo;
	UCHAR	*pReadAdr;
	UCHAR	i;
	
	/* Parameter Check */
	if( PARAM_OK == FALSE ){										/* parameter error? */
		return( ERRPRM );											/* then rerun */
	}
	
	/* Searches for valid blocks currently used as EEPROM */
	ucBlockTBLNo = ucEEPROMUseBlockSearchEx();
	if ( ucBlockTBLNo == FALSE ){									/* abnormal end? */
		return ( ERRNOUSE );										/* With no effective block */
	}
	/* Obtains address of latest data for specified number */
	usDataSrcAdr = usEEPROMDataSearchEx( ucDataNo, ucBlockTBLNo );
	if ( usDataSrcAdr == 0x00 ){
		return( ERRNODAT );											/* no data */
	}
	/* Reads latest data from obtained address */
	pReadAdr = (UCHAR*)( usDataSrcAdr +1 );
	for ( i=0 ; i < DATALENGTH ; i++ ){
		*pDataAddress++ = *pReadAdr++;
	}
	if ( ( (USHORT)pReadAdr & 0x0003 ) == 0x0003 ){					/* Offset is +3 located (a 4byte unit) */
		/* Since the delimiter is read, it rewrites by the last byte data */
		pDataAddress--;												/* One pointer is returned */
		*pDataAddress = *pReadAdr;
	}
	return( TRUE );
}

/*-------------------------------------------------------------------
 Function name : ucEEPROMWriteEx(user access function)
 Arguments	   : ucDataNo:Write data number
			   : *pDataAddress:Write data storage address
 Return value  : Normal completion(TRUE:00H) / Data full(ERRFULL:F5H) / Write error(ERRWRITE:F6H) /
                 With no effective block(ERRNOUSE:FAH) / Initialization error(ERRINIT:FEH)
 Summary       : Writes data with specified number from the storage address to EEPROM.
			   : 1.Searches for blocks being used as EEPROM.
			   : 2:Searches for addresses to enable writing to valid block.
			   : 3:Creates write data.
			   : 4:Writes to valid blocks.
-------------------------------------------------------------------*/
UCHAR ucEEPROMWriteEx( UCHAR ucDataNo, UCHAR *pDataAddress )		/* EEPROM data write processing */
{
	USHORT	usBlockAdr;
	USHORT	usWriteAdr;
	UCHAR	ucBlockTBLNo;
	UCHAR	ucCount;
	UCHAR	ucResult;
	UCHAR	i;
	
	/* Parameter Check */
	if( PARAM_OK == FALSE ){										/* parameter error? */
		return( ERRPRM );											/* then rerun */
	}
	
	/* Searches for valid blocks being used as EEPROM */
	ucBlockTBLNo = ucEEPROMUseBlockSearchEx();
	if ( ucBlockTBLNo == FALSE ){									/* abnormal end */
		return( ERRNOUSE );											/* With no effective block */
	}
	
	/* Bank set */
	SET_BANK( ucEEPROM_BANK[ ucBlockTBLNo ] );
	/* Bank final address acquisition */
	usBlockAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ ucBlockTBLNo ] );
	usBlockAdr += DATA_END_OFFSET;
	/* Obtains EEPROM write address */
	usWriteAdr = usEEPROMWriteTopSearchEx( ucBlockTBLNo );
	/* Uses data specified by DataNo to create data to write to EEPROM */
	ucDataBuf[ WRITE_DATA_LENGTH    ] = 0xFF;						/* Last field: Delimiter */
	ucDataBuf[ WRITE_DATA_LENGTH -1 ] = 0xFF;
	ucDataBuf[ WRITE_DATA_LENGTH -2 ] = 0xFF;
	ucDataBuf[ WRITE_DATA_LENGTH -3 ] = 0xFF;						/* FFH setting of empty data */
	ucDataBuf[ 0 ] = ucDataNo;										/* Top: Data number	*/
	for ( i = 1 ; i <= DATALENGTH ; i++ ){
		ucDataBuf[ i ] = *pDataAddress++;							/* Setting of writing data */
	}
	/* A delimiter is set */
	ucDataBuf[ WRITE_DATA_LENGTH   ] = ucDataBuf[ WRITE_DATA_LENGTH-1 ];	/* The last data is moved to the last byte */
	ucDataBuf[ WRITE_DATA_LENGTH-1 ] = 0x00;
	
	SelfFlashModeOn();												/* changes to a self-programming mode */
	if ( ucSelfFlashInitialize() == FALSE ){						/* self-programming Initialization */
		SelfFlashModeOff();											/* changes to a normal mode */
		return( ERRINIT );											/* initialization error */
	}
	/* Write-in counter initialization */
	ucCount = WRITENUM + 1;
	while( 1 ) {													/* loop */
		/* An empty check */
		if ( usWriteAdr >= usBlockAdr ) {							/* With no opening */
			ucResult = ERRFULL;										/* Data full */
			break;
		}
		/* Writes write data to EEPROM */
		ucResult = ucSelfFlashEEPROMWrite( ucEEPROM_BANK[ ucBlockTBLNo ], usWriteAdr, WRITE_DATA_LENGTH );
		if ( ucResult == TRUE ){									/* Normal end? */
			break;
		}
		else if ( ucResult != LIB_BLANK_ERR ){						/* Except a blank error */
			ucCount--;												/* Counter-1 */
			if ( ucCount <= 0 ) {									/* retry over */
				ucResult = ERRWRITE;								/* write error */
				break;
			}
		}
		/* Address update */
		usWriteAdr += WRITE_DATA_LENGTH+1;
	}
	/* changes to a normal mode */
	SelfFlashModeOff();
	return ( ucResult );
}

/*-------------------------------------------------------------------
 Function name : ucEEPROMChangeEx
 Arguments	   : None
 Return value  : Normal completion(TRUE:00H) / Noemal end(Block eror)(NMLBLK:01H) / 
                 Block nothing finished with elimination(ERRNOBLA:F8H) /
                 With no effective block(ERRNOUSE:FAH) / Device error(ERRDEV:FDH) /
                 Initialization error(ERRINIT:FEH)
 Summary       : If the currently used blocks are full of data, this function searches for the next
 			   : block to be used and copies data to that block.
			   : 1.Sets block to be used next.
			   : 2.Erases block to be used next.
			   : 3.Transfers the latest data from a valid block to the next block.
			   : 4.Sets the next block to be used as valid.
			   : 5.Sets currently valid blocks as invalid.
-------------------------------------------------------------------*/
UCHAR ucEEPROMChangeEx( void )										/* EEPROM block change processing */
{
	UCHAR	ucFlags;			/* Various flags          */
								/*   D7:block error       */
								/*   D6:Block elimination */
								/*   D5:not use           */
								/*   D4:not use           */
								/*   D3:not use           */
								/*   D2:not use           */
								/*   D1:not use           */
								/*   D0:not use           */
	UCHAR	ucCount;			/* erase counter          */
	USHORT	usOldAdr;			/* old block address      */
	USHORT	usNewAdr;			/* new block address      */
	UCHAR	*pGetAdr;			/* Read address           */
	UCHAR	*pSetAdr;			/* Write address          */
	UCHAR	ucOldBlockTBLNo;	/* old block no.          */
	UCHAR	ucNewBlockTBLNo;	/* new block no.          */
	UCHAR 	i;
	UCHAR 	j;
	USHORT 	k;
	
	/* Parameter Check */
	if( PARAM_OK == FALSE ){										/* parameter error? */
		return( ERRPRM );											/* then rerun */
	}
	
	/* Flag initialization */
	ucFlags = 0x00;
	/* Searches for valid blocks being used as EEPROM */
	ucOldBlockTBLNo = ucEEPROMUseBlockSearchEx();
	if ( ucOldBlockTBLNo == FALSE ){
		return ( ERRNOUSE );										/* With no effective block */
	}
	/* Bank set */
	SET_BANK( ucEEPROM_BANK[ ucOldBlockTBLNo ] );
	/* bank top address(flag address) get */
	usOldAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ ucOldBlockTBLNo ] );
	/* processing block no. set */
	ucNewBlockTBLNo = ucOldBlockTBLNo;
	/* block change process */
	for ( i = 0 ; i < EEPROM_BLOCK_NO ; i++ ) {						/* loop */
		/* block no. update */
		if ( ucNewBlockTBLNo >= (EEPROM_BLOCK_NO-1) ){				/* last block? */
			ucNewBlockTBLNo = 0;									/* returns to a head */
		}
		else {
			ucNewBlockTBLNo++;										/* block No+1 */
		}
		/* Bank set */
		SET_BANK( ucEEPROM_BANK[ ucNewBlockTBLNo ] );
		/* bank top address(flag address) get */
		usNewAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ ucNewBlockTBLNo ] );
		/* effective/Invalid flag check */
		if( BLOCK_IS_PROHIBIT( (UCHAR*)usNewAdr ) == FALSE ){		/* blockExcept prohibition of use? */
			
#ifdef AUTOERASE
			SelfFlashModeOn();										/* changes to a self-programming mode */
			if ( ucSelfFlashInitialize() != TRUE ){					/* self-programming Initialization */
				SelfFlashModeOff();	/* changes to a normal mode */
				return ( ERRINIT );	/* initialization error */
			}
			for ( ucCount = ERASENUM + 1; ucCount > 0; ucCount-- ){	/* loop */
				/* Erases data in the block next used as EEPROM */
				if ( ucSelfFlashBlockErase( ucEEPROM_BANK[ ucNewBlockTBLNo ], ucEEPROM_BLOCK[ ucNewBlockTBLNo ] ) == TRUE ){	/* Normal end */
					break;
				}
			}
			if ( ucCount <= 0 ){									/* counter <= 0 (retry over) */
				/* block is made prohibition of use */
				if ( ucEEPROMUnableEx( ucNewBlockTBLNo ) == FALSE ){/* abnormal end */
					SelfFlashModeOff();								/* changes to a normal mode */
					return ( ERRDEV );								/* device error */
				}
				ucFlags |= FL_BLKERR;								/* Block error flag On */
				continue;
			}
			SelfFlashModeOff();										/* changes to a normal mode */
#endif	/* AUTOERASE */
				
			if( BLOCK_IS_BLANK( (UCHAR*)usNewAdr ) == TRUE ){		/* blockstate of eliminating? */
				/* Sets write start address for next block's data */
				pSetAdr = (UCHAR *)( usNewAdr + DATATOP ) ;
				/* Transfers latest EEPROM data from current block to next block */
				for ( j = 0 ; j < DATA_NO_MAX ; j++ ){
					/* The newest data search */
					pGetAdr = (UCHAR *)usEEPROMDataSearchEx( j, ucOldBlockTBLNo );
					if ( (USHORT)pGetAdr != 0 ){					/* Those with data? */
						for ( k=0 ; k < (WRITE_DATA_LENGTH+1) ; k++ ){
							ucDataBuf[k] = *pGetAdr++;
						}
						SET_BANK( ucEEPROM_BANK[ ucNewBlockTBLNo ] );/* Bank set */
						SelfFlashModeOn();							/* changes to a self-programming mode */
						if ( ucSelfFlashInitialize() != TRUE ){		/* self-programming Initialization */
							SelfFlashModeOff();						/* changes to a normal mode */
							return ( ERRINIT );						/* initialization error */
						}
						for ( ucCount = (WRITENUM + 1); ucCount > 0; ucCount--){
							/* write to new block */
							if ( ucSelfFlashEEPROMWrite( ucEEPROM_BANK[ ucNewBlockTBLNo ], (USHORT)pSetAdr, WRITE_DATA_LENGTH ) == TRUE ){
								break;
							}
							/* Address update */
							pSetAdr += WRITE_DATA_LENGTH+1;
						}
						if ( ucCount <= 0 ) {						/* write error */
							/* Except prohibition of use */
							if ( ucEEPROMUnableEx( ucNewBlockTBLNo ) == FALSE ){	/* abnormal end */
								SelfFlashModeOff();					/* changes to a normal mode */
								return ( ERRDEV );					/* device error */
							}
							SelfFlashModeOff();						/* changes to a normal mode */
							ucFlags |= FL_BLKERR;					/* Block error flag On */
							break;
						}
						SelfFlashModeOff();							/* changes to a normal mode */
						/* őf[^Address update */
						pSetAdr += WRITE_DATA_LENGTH+1;
					}
				}
				/* Renewal processing of a flag */
				if( j >= DATA_NO_MAX ) {							/* data copy Normal end? */
					/* write 55555555h */
					ucDataBuf[ 0 ] = 0x55;
					ucDataBuf[ 1 ] = 0x55;
					ucDataBuf[ 2 ] = 0x55;
					ucDataBuf[ 3 ] = 0x55;
					SelfFlashModeOn();								/* changes to a self-programming mode */
					if ( ucSelfFlashInitialize() != TRUE ){			/* self-programming Initialization */
						SelfFlashModeOff();							/* changes to a normal mode */
						return ( ERRINIT );							/* initialization error */
					}
					/* new block is effective  */
					if ( ucSelfFlashEEPROMWrite( ucEEPROM_BANK[ ucNewBlockTBLNo ], usNewAdr, 4-1 ) == TRUE ) {	/* Normal end */
						/* block is Invalid */
						if ( ucSelfFlashEEPROMWrite( ucEEPROM_BANK[ ucOldBlockTBLNo ], (usOldAdr+4), 4-1 ) != TRUE ) {	/* abnormal end */
							/* write errormake a block the ban on use */
							if ( ucEEPROMUnableEx( ucOldBlockTBLNo ) == FALSE ){	/* abnormal end */
								SelfFlashModeOff();					/* changes to a normal mode */
								return ( ERRDEV );					/* device error */
							}
							ucFlags |= FL_BLKERR;					/* Block error flag On */
						}
						SelfFlashModeOff();							/* changes to a normal mode */
						break;
					}
					else {											/* abnormal end */
						/* write errormake a block the ban on use */
						if ( ucEEPROMUnableEx( ucNewBlockTBLNo ) == FALSE ){	/* abnormal end */
							SelfFlashModeOff();						/* changes to a normal mode */
							return ( ERRDEV );						/* device error */
						}
						ucFlags |= FL_BLKERR;						/* Block error flag On */
					}
					SelfFlashModeOff();								/* changes to a normal mode */
				}
			}
		}
	}
	/* With no eliminated block */
	if ( i >= EEPROM_BLOCK_NO ) {
		return ( ERRNOBLA );										/* With no eliminated block */
	}
	if ( ucFlags & FL_BLKERR ) {									/* Those with a block error? */
		return ( NMLBLK );											/* Normal end(Block error generating) */
	}
	return ( TRUE );												/* Normal end */
}

/*-------------------------------------------------------------------
 Function name : ucEEPROMEraseEx (user access function)
 Arguments     : None
 Return value  : Normal completion(TRUE:00H) / With no eliminable block(ERRNOERS:F9H) /
                 With no effective block(ERRNOUSE:FAH) / Block eror(ERRBLK:FCH) /
                 Device error(ERRDEV:FDH) / Initialization error(ERRINIT:FEH)
 Summary       : Erases data in the block to which data is written to next in advance
-------------------------------------------------------------------*/
UCHAR ucEEPROMEraseEx( void ) 										/* EEPROM erase processing */
{
	UCHAR	ucBlockTBLNo;		/* erase block no    */
	UCHAR	ucCount;			/* erase Counter     */
	UCHAR	ucResult;			/* Processing result */
	USHORT	usFlagAdr;			/* flag address      */
	UCHAR	i;
	
	/* Parameter Check */
	if( PARAM_OK == FALSE ){										/* parameter error? */
		return( ERRPRM );											/* then rerun */
	}
	
	/* Searches for valid blocks being used as EEPROM */
	ucBlockTBLNo = ucEEPROMUseBlockSearchEx();
	if ( ucBlockTBLNo == FALSE ){									/* abnormal end */
		return ( ERRNOUSE );										/* With no effective block */
	}
	/* Processing result initial value */
	ucResult = TRUE;
	/* The loop for several block minutes is performed(Except an effective block) */
	for ( i = 0 ; i < (EEPROM_BLOCK_NO - 1) ; i++ ) {
		/* blockNo.+P */
		if ( ucBlockTBLNo <= 0 ){
			ucBlockTBLNo = EEPROM_BLOCK_NO - 1;						/* last block No. */
		}
		else {
			ucBlockTBLNo--;											/* blockNo.-1 */
		}
		/* Bank set */
		SET_BANK( ucEEPROM_BANK[ ucBlockTBLNo ] );
		/* bank top address(flag address) get */
		usFlagAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ ucBlockTBLNo ] );
		/* effective/Invalid flag check */
		if( BLOCK_IS_PROHIBIT( (UCHAR*)usFlagAdr ) == FALSE ){		/* Except prohibition of use? */
			/* erase process */
			SelfFlashModeOn();										/* changes to a self-programming mode */
			if ( ucSelfFlashInitialize() != TRUE ){					/* self-programming Initialization */
				SelfFlashModeOff();									/* changes to a normal mode */
				return ( ERRINIT );									/* initialization error */
			}
			for ( ucCount = (ERASENUM + 1); ucCount > 0; ucCount-- ){	/* loop */
				/* Erases data in the block next used as EEPROM */
				if ( ucSelfFlashBlockErase( ucEEPROM_BANK[ ucBlockTBLNo ], ucEEPROM_BLOCK[ ucBlockTBLNo ] ) == TRUE ){	/* Normal end */
					break;
				}
			}
			if ( ucCount <= 0 ){									/* counter <= 0 (retry over) */
				/* Except prohibition of use */
				if ( ucEEPROMUnableEx( ucBlockTBLNo ) == FALSE ){	/* abnormal end */
					ucResult = ERRDEV;								/* device error */
				}
				else {
					ucResult = ERRBLK;								/* block error */
				}
			}
			SelfFlashModeOff();										/* changes to a normal mode */
			break;
		}
	}
	/* Processing result */
	if ( i >= (EEPROM_BLOCK_NO - 1) ) {								/* With no eliminable block */
		ucResult = ERRNOERS;										/* return valueWith no eliminable block */
	}
	return ( ucResult );
}

/*-------------------------------------------------------------------
 Function name : ucEEPROMGetRemainEx (user access function)
 Arguments     : None
 Return value  : The number of data which can be written in
 Summary       : examine the number of the writable data
-------------------------------------------------------------------*/
UCHAR ucEEPROMGetRemainEx( void )
{
	UCHAR	ucBlockTBLNo;	/* erase blockNo.    */
	UCHAR	ucResult;		/* Processing result */
	USHORT	usWriteAdr;		/* write top address */
	USHORT	usBlockAdr;		/* block擪address  */
	
	/* Parameter Check */
	if( PARAM_OK == FALSE ){										/* parameter error? */
		return( FALSE );											/* then rerun */
	}
	
	ucResult = 0;													/* initial value */
	
	/* Searches for valid blocks being used as EEPROM */
	ucBlockTBLNo = ucEEPROMUseBlockSearchEx();
	if ( ucBlockTBLNo == FALSE ){									/* With no effective block? */
		return ( FALSE );											/* With no effective block */
	}
	/* Bank set */
	SET_BANK( ucEEPROM_BANK[ ucBlockTBLNo ] );
	/* top address */
	usBlockAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ ucBlockTBLNo ] );	/* Obtains EEPROM write address */
	/* write address */
	usWriteAdr = usEEPROMWriteTopSearchEx( ucBlockTBLNo );
	/* The number calculation of data */
	ucResult = (UCHAR)((DATA_END_OFFSET - (usWriteAdr - usBlockAdr)) / (WRITE_DATA_LENGTH+1));
	return ( ucResult );
}

/*-------------------------------------------------------------------

/*-------------------------------------------------------------------
 Function name : ucEEPROMUseBlockSearchEx
 Arguments	   : None
 Return value  : Block table number(0x00 to 0xFE), Abnormal completion(0xFF)
 Summary       : Searches for currently used blocks in flash memory that are allocated to EEPROM.
-------------------------------------------------------------------*/
UCHAR ucEEPROMUseBlockSearchEx( void )								/* EEPROM block search processing */
{
	UCHAR	i;
	USHORT	usFlagAdr;
	
	/* Searches for currently used block (block with a valid "valid/invalid" flag) */
	for ( i = 0 ; i < EEPROM_BLOCK_NO ; i++ ){
		/* Bank set */
		SET_BANK( ucEEPROM_BANK[ i ] );
		/* abnk top address */
		usFlagAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ i ] );
		if ( BLOCK_IS_EFFECTIVE( (UCHAR*)usFlagAdr ) == TRUE ){		/* blockeffective? */
			return( i );											/* the block is being used */
		}
	}
	return( FALSE );
}

/*-------------------------------------------------------------------
 Function name : usEEPROMWriteTopSearchEx
 Arguments	   : ucBlockTBLNo:Currently used block table number(block number of block set for
 			   : EEPROM use by user definition)
 Return value  : Write address(0x0001 to 0xffff), Abnormal completion(0x0000)
 Summary       : Searches for specified block's write address.
-------------------------------------------------------------------*/
USHORT usEEPROMWriteTopSearchEx( UCHAR ucBlockTBLNo )				/* EEPROM block data write top search processing */
{
	USHORT	usTopAdr;
	UCHAR	*pSearch;
	UCHAR	*pWrite;
	
	/* Bank set */
	SET_BANK( ucEEPROM_BANK[ ucBlockTBLNo ] );
	/* bank top address */
	usTopAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ ucBlockTBLNo ] );	/* Obtains EEPROM write address */
	/* address search & set write address */
	pSearch = pWrite = (UCHAR *)( usTopAdr + DATA_END_OFFSET );
	/* set a data address */
	usTopAdr += DATATOP;
	
	while ( 1 ){
		/* search address - data length */
		pSearch -= WRITE_DATA_LENGTH+1;
		if ( (USHORT)pSearch <  usTopAdr ){							/* search finish? */
			break;
		}
		/* check delimiter */
		if ( *(pSearch + (WRITE_DATA_LENGTH-1)) != 0xFF ){			/* FFh? */
			break;
		}
		/* save address */
		pWrite = pSearch;
	}
	return ( (USHORT)pWrite );
}

/*-------------------------------------------------------------------
 Function name : usEEPROMDataSearchEx
 Arguments	   : ucData:Read data No.
 		       : ucBlockTBLNo:Currently used block table number(block number of block set for
 		       : EEPROM use by user definition)
 Return value  : Address of latest data(0x0001 to 0xffff), Abnormal completion(0x0000)
 Summary       : Reads storage address of latest data at specified number.
-------------------------------------------------------------------*/
USHORT usEEPROMDataSearchEx( UCHAR ucDataNo, UCHAR ucBlockTBLNo )	/* Processing to read latest specified data from EEPROM blocks */
{
	USHORT	usTopAdr;
	UCHAR	*pSearch;
	
	/* Bank set */
	SET_BANK( ucEEPROM_BANK[ ucBlockTBLNo ] );
	/* bank top address */
	usTopAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ ucBlockTBLNo ] );	/* Obtains EEPROM write address */
	/* search address */
	pSearch = (UCHAR *)( usTopAdr + DATA_END_OFFSET );
	/* data address */
	usTopAdr += DATATOP;
	
	while ( 1 ){
		/* search address - data length */
		pSearch -= WRITE_DATA_LENGTH+1;
		if ( (USHORT)pSearch < usTopAdr ){							/* finish? */
			pSearch = 0;											/* set 0 */
			break;
		}
		/* check delimiter */
		if ( *(pSearch + (WRITE_DATA_LENGTH-1)) == 0x00 ){			/* 00h? */
			/* check data No. */
			if ( *pSearch == ucDataNo ){							/* coincidence? */
				break;
			}
		}
	}
	return ( (USHORT)pSearch );
}

/*-------------------------------------------------------------------
 Function name : ucEEPROMUnableEx
 Arguments	   : ucBlockTBLNo:Block table number(0x00 to 0xFE)
 Return value  : Normal completion(TRUE:0x00) / Abnormal completion(FALSE:0xff)
 Summary       : make a block the ban on use
-------------------------------------------------------------------*/
UCHAR ucEEPROMUnableEx( UCHAR ucBlockTBLNo )		/* EEPROM latest data recover processing */
{
	USHORT	usSetAdr;
	UCHAR	ucCount;
	UCHAR	ucResult;
	UCHAR	i;
	
	ucResult = FALSE;
	SET_BANK( ucEEPROM_BANK[ ucBlockTBLNo ] );
	usSetAdr = BLOCK_TO_ADDRESS( ucEEPROM_BLOCK[ ucBlockTBLNo ] );
	/* effective/Invalid flag */
	for ( i = 0 ; i < 4 ; i++ ){
		ucDataBuf[ i ] = 0x00;
	}
	/* effective flag update */
	for ( ucCount = (WRITENUM + 1); ucCount > 0; ucCount-- ){
		if( ucSelfFlashWordWrite( ucEEPROM_BANK[ ucBlockTBLNo ], usSetAdr, (USHORT)4 ) == TRUE ) {
			ucResult = TRUE;
			break;
		}
	}
	usSetAdr += 4;
	/* Invalid flag update */
	for ( ucCount = (WRITENUM + 1); ucCount > 0; ucCount-- ){
		if( ucSelfFlashWordWrite( ucEEPROM_BANK[ ucBlockTBLNo ], usSetAdr, (USHORT)4 ) == TRUE ) {
			ucResult = TRUE;
			break;
		}
	}
	return( ucResult );
}

/*-------------------------------------------------------------------

/*-------------------------------------------------------------------
 Function name : SelfFlashModeOn
 Arguments	   : None
 Return value  : None
 Summary       : Sets self rewriting mode(A1).
-------------------------------------------------------------------*/
void SelfFlashModeOn( void )										/* Processing to switch from normal mode to self programming mode */
{
	DI();
	FlashStart();													/* Self programing start library call */
	EI();
	return;
}

/*-------------------------------------------------------------------
 Function name : SelfFlashModeOff
 Arguments	   : None
 Return value  : None
 Summary       : Releases self rewriting mode.
-------------------------------------------------------------------*/
void SelfFlashModeOff( void )										/* Processing to switch from self programming mode to normal mode */
{
	DI();
	FlashEnd();														/* Self programing end library call */
	EI();
	return;
}

/*-------------------------------------------------------------------
 Function name : ucSelfFlashInitialize
 Arguments	   : None
 Return value  : Normal completion(TRUE:0x00) / Abnormal completion(FALSE:0xff)
 Summary       : Initializes self programming firmware for EEPROM emulation.
			   : Initializes flash firmware (sets operating frequency)
			   : Checks FLMD pin [because programming may be disabled depending on the FLMD pin's status]
-------------------------------------------------------------------*/
UCHAR ucSelfFlashInitialize( void )									/* Flash memory self programming initialization processing */
{
	FlashEnv( (USHORT)&ucEntryRAM );								/* Initializes internal flash firmware */
	
	if( CheckFLMD() != LIB_NORMAL_END ){							/* Obtains FLMD pin information */
		return( FALSE );
	}
	return( TRUE );
}

/*-------------------------------------------------------------------
 Function name : ucSelfFlashBlockErase
 Arguments	   : ucEraseStartBlock:Erase block number
 Return value  : Normal completion(TRUE:0x00) / Abnormal completion(FALSE:0xFF)
 Summary       : Erases specified block.
-------------------------------------------------------------------*/
UCHAR ucSelfFlashBlockErase( UCHAR ucBank, UCHAR ucBlock )			/* Flash memory block erase processing */
{
	UCHAR	ucStatus;
	
	while(1){
		DI();
		ucStatus = FlashBlockBlankCheck( ucBank, ucBlock );			/* Flash memory blank check library call */
		EI();
		if ( ucStatus == LIB_BLKCHK_ERR ){							/* Blank check error (Blank check error) */
			DI();
			ucStatus = FlashBlockErase( ucBank, ucBlock );			/* Flash memorry erase library call */
			EI();
			if (( ucStatus != LIB_NORMAL_END )&&					/* States other than discontinuation error and nomarl error */
				( ucStatus != LIB_STATE_INT )){						/* (parameter/protect/erase error) */
				return( FALSE );
			}
		}
		else if ( ucStatus == LIB_NORMAL_END ){						/* Blank check normal end */
			FlashEnv( (USHORT)&ucEntryRAM );
			return( TRUE );
		}
		else if ( ucStatus != LIB_STATE_INT ){						/* States other than discontinuation error (parameter error)*/
			return( FALSE );
		}
	}
}

/*-------------------------------------------------------------------
 Function name : ucSelfFlashEEPROMWrite
 Arguments     : ucBank:Bank number
               : ucNum:Data number
               : usAddr:Writes start address
 Return value  : Normal completion(TRUE:0x00) / Abnormal completion(FALSE:0xFF)
 Summary       : write data to EEPROM
-------------------------------------------------------------------*/
UCHAR ucSelfFlashEEPROMWrite( UCHAR ucBank, USHORT usAddr, UCHAR ucNum )	/* Flash memory data write processing */
{
	UCHAR	ucStatus;
	
	WordAddr.WriteAddress = usAddr;
	WordAddr.WriteBank = ucBank;
	
	while(1){
		DI();
		ucStatus = EEPROMWrite( &WordAddr, (UCHAR)((ucNum/4)+1), (USHORT)( &ucDataBuf[ 0 ] ));	/* Flash memory EEPROM write library call */
		EI();
		if ( ucStatus == LIB_NORMAL_END ){							/* Word write normal end */
			return( TRUE );
		}
		else if ( ucStatus != LIB_STATE_INT ){						/* States other than discontinuation error (parameter/protect/write error) */
			return( ucStatus );
		}
	}
}

/*-------------------------------------------------------------------
 Function name : ucSelfFlashWordWrite
 Arguments     : usSetAdr:
               : usDataAdr:
 Return value  : Normal completion(TRUE:0x00) / Abnormal completion(FALSE:0xFF)
 Summary       : write word data to EEPROM
-------------------------------------------------------------------*/
UCHAR ucSelfFlashWordWrite( UCHAR ucBank, USHORT usSetAdr, UCHAR ucNum )	/* Flash memory word data write processing */
{
	UCHAR	ucStatus;
	
	WordAddr.WriteAddress = usSetAdr;
	WordAddr.WriteBank = ucBank;

	while(1){
		DI();
		ucStatus = FlashWordWrite( &WordAddr, (UCHAR)((((USHORT)ucNum-1)/4)+1), (USHORT)(&ucDataBuf[0]) );	/* Word write library call */
		EI();
		if ( ucStatus == LIB_NORMAL_END ){							/* Word write normal end */
			return( TRUE );
		}
		else if ( ucStatus != LIB_STATE_INT ){						/* States other than discontinuation error (parameter/protect/write error) */
			return( FALSE );
		}
	}
}

/*-------------------------------------------------------------------

/*-------------------------------------------------------------------
 Function name : usBlockToAddress
 Arguments	   : ucBlock:Block number
 Return value  : Address Data
 Summary       : A block number is changed into an address.
-------------------------------------------------------------------*/
USHORT usBlockToAddress( USHORT usBlock )
{
	return ( (usBlock << 2) * 0x0100 );
}

/*-------------------------------------------------------------------
 Function name : ucUseProhibitionBlock
 Arguments	   : usFlagAdr:Flag Address
 Return value  : 00h(TRUE) / FFh(FALSE)
 Summary       : examine it whether it is a ban on use block
-------------------------------------------------------------------*/
UCHAR ucUseProhibitionBlock( UCHAR *p )
{
	if (((p[0]==0x00)&&(p[1]==0x00)&&(p[2]==0x00)&&(p[3]==0x00))||
		((p[4]==0x00)&&(p[5]==0x00)&&(p[6]==0x00)&&(p[7]==0x00))){
		return ( TRUE );
	}
	return ( FALSE );
}

/*-------------------------------------------------------------------
 Function name : ucEffectiveBlock
 Arguments	   : usFlagAdr:Flag Address
 Return value  : 00h(TRUE) / FFh(FALSE)
 Summary       : examine whether I block it effectively
-------------------------------------------------------------------*/
UCHAR ucEffectiveBlock( UCHAR *p )
{
	if (((p[0]==0x55)&&(p[1]==0x55)&&(p[2]==0x55)&&(p[3]==0x55))&&
		((p[4]==0xFF)&&(p[5]==0xFF)&&(p[6]==0xFF)&&(p[7]==0xFF))){
		return ( TRUE );
	}
	return ( FALSE );
}

/*-------------------------------------------------------------------
 Function name : ucBlankBlock
 Arguments	   : usFlagAdr:Flag Address
 Return value  : 00h(TRUE) / FFh(FALSE)
 Summary       : examine it whether it is a blank block
-------------------------------------------------------------------*/
UCHAR ucBlankBlock( UCHAR *p )
{
	if (((p[0]==0xFF)&&(p[1]==0xFF)&&(p[2]==0xFF)&&(p[3]==0xFF))&&
		((p[4]==0xFF)&&(p[5]==0xFF)&&(p[6]==0xFF)&&(p[7]==0xFF))){
		return ( TRUE );
	}
	return ( FALSE );
}

/*-------------------------------------------------------------------
 Function name : ucParamCheck
 Arguments	   : none
 Return value  : 00h(TRUE) / FFh(FALSE)
 Summary       : check parameteres
-------------------------------------------------------------------*/
UCHAR ucParamCheck( void )
{
	if(( DATALENGTH < 1) || (DATALENGTH > 254)){
		return( FALSE );
	}
	if(DATANOMAX > 253){
		return( FALSE );
	}
	if(WRITENUM > 253){
		return( FALSE );
	}
	if(ERASENUM > 254){
		return( FALSE );
	}
	return( TRUE );
}

